import collections
from multiprocessing.sharedctypes import Value

import numpy as np

FLOAT_32_LE = np.dtype("<f4")
RSC_KEY_TYPES = [
    ("width", int),
    ("file_length", int),
    ("x_first", float),
    ("y_first", float),
    ("x_step", float),
    ("y_step", float),
    ("x_unit", str),
    ("y_unit", str),
    ("z_offset", int),
    ("z_scale", int),
    ("projection", str),
]


def load(
    filename,
    rsc_file=None,
    rows=None,
    cols=None,
    band=1,
    mask_nodata=True,
    **kwargs,
):
    """Load a file, either using numpy or rasterio

    Parameters
    ----------
    filename : str
        Name of file to load
    rsc_file : int
         (Default value = None)
    rows : int
        Manually specify number of rows of image
    cols :
        Manually specify number of cols of image
    band : int
        For gdal, specify the band of the image to load (Default value = 1)
    mask_nodata : bool
        If True, convert out gdal-indicated nodata values as nans (Default value = True)
    **kwargs :
        

    Returns
    -------
    ndarray : image data
    """
    if rsc_file:
        rsc_data = load_rsc(rsc_file)
        return load_stacked_img(filename, rsc_data=rsc_data, rows=rows, cols=cols)
    else:
        try:
            from osgeo import gdal
            gdal.UseExceptions()
        except ImportError:
            raise ValueError("Need to `conda install gdal` to load gdal-readable")

        ds = gdal.Open(filename)
        image = ds.GetRasterBand(band).ReadAsArray()
        # Get the nodata value, and set it to nan
        nodata = ds.GetRasterBand(band).GetNoDataValue()
        if nodata is not None:
            try:
                image[image == nodata] = np.nan
            except ValueError:  # non-float image type, dont mask
                pass
        ds = None
        return image


def load_rsc(filename, lower=False, **kwargs):
    """Loads and parses the .rsc file

    Parameters
    ----------
    lower : bool
        make keys of the dict lowercase (Default value = False)
    filename :
        
    **kwargs :
        

    Returns
    -------
    dict
        .rsc file parsed out
        example file:
        WIDTH         10801
        FILE_LENGTH   7201
        X_FIRST       -157.0
        Y_FIRST       21.0
        X_STEP        0.000277777777
        Y_STEP        -0.000277777777
        X_UNIT        degrees
        Y_UNIT        degrees
        Z_OFFSET      0
        Z_SCALE       1
        PROJECTION    LL

    """

    # Use OrderedDict so that upsample_dem_rsc creates with same ordering as old
    output_data = collections.OrderedDict()
    # Second part in tuple is used to cast string to correct type

    with open(filename, "r") as f:
        for line in f.readlines():
            for field, num_type in RSC_KEY_TYPES:
                if line.startswith(field.upper()):
                    output_data[field] = num_type(line.split()[1])

    if lower:
        output_data = {k.lower(): d for k, d in output_data.items()}
    return output_data


def load_stacked_img(
    filename,
    rows=None,
    cols=None,
    rsc_data=None,
    return_amp=False,
    dtype=FLOAT_32_LE,
    **kwargs,
):
    """Helper function to load .unw and .cor files from snaphu output
    
    Format is two stacked matrices:
        [[first], [second]] where the first "cols" number of floats
        are the first matrix, next "cols" are second, etc.
    Also called BIL, Band Interleaved by Line
    See http://webhelp.esri.com/arcgisdesktop/9.3/index.cfm?topicname=BIL,_BIP,_and_BSQ_raster_files
    for explantion
    
    For .unw height files, the first is amplitude, second is phase (unwrapped)
    For .cc correlation files, first is amp, second is correlation (0 to 1)

    Parameters
    ----------
    filename :
        str
    rows :
        int (Default value = None)
    cols :
        int (Default value = None)
    rsc_data :
        dict (Default value = None)
    return_amp :
        bool (Default value = False)
    dtype :
         (Default value = FLOAT_32_LE)
    **kwargs :
        

    Returns
    -------
    type
        ndarray: dtype=float32, the second matrix (height, correlation, ...) parsed
        if return_amp == True, returns two ndarrays stacked along axis=0

    """
    if rows is None or cols is None:
        rows, cols = rsc_data["file_length"], rsc_data["width"]

    data = np.fromfile(filename, dtype)

    first = data.reshape((rows, 2 * cols))[:, :cols]
    second = data.reshape((rows, 2 * cols))[:, cols:]
    if return_amp:
        return np.stack((first, second), axis=0)
    else:
        return second


def load_mask(mask_files, rsc_file=None, mask_is_zero=False):
    """Create one mask from a list of mask files

    Parameters
    ----------
    mask_files :
        
    rsc_file :
         (Default value = None)

    Returns
    -------
    ndarray
        dtype=bool, the mask

    """
    mask = np.ma.nomask
    for mask_file in mask_files:
        cur_mask = load(mask_file, band=1, rsc_file=rsc_file).astype(np.bool)
        if mask_is_zero:
            cur_mask = ~cur_mask
        mask = np.logical_or(mask, cur_mask)
    return mask



import argparse

from . import core

description = """
'--level pixel' means individual pixels for each SAR date are labeled (good for larger scenes).
'--level scene' means whole SAR images are labeled (good for smaller scenes).
For scene level labeling, the variance of each average interferogram is used.
"""


def get_cli_args():
    """ """
    p = argparse.ArgumentParser(description=description)
    p.add_argument(
        "--input-files",
        "-i",
        help=(
            "Provide the list of input interferogram filenames in a text file."
            "Alternate to `--search-path`."
        ),
    )
    p.add_argument(
        "--search-path",
        "-p",
        default=".",
        help="location of igram files. (default=%(default)s)",
    )
    p.add_argument(
        "--ext",
        default=".unw",
        help="filename extension of unwrapped igrams to average (default=%(default)s)",
    )
    p.add_argument(
        "--band",
        type=int,
        default=1,
        help="If using GDAL to load igrams, specify which band represents unwrapped phase (default=%(default)s)",
    )
    p.add_argument(
        "--outfile",
        "-o",
        default="labels.nc",
        help="Location to save final labels (default=%(default)s)",
    )
    p.add_argument(
        "--level",
        default="scene",
        choices=["pixel", "scene"],
        help=("Level at which to label outliers. (default=%(default)s).\n"),
    )
    p.add_argument(
        "--avg-file",
        default="average_ifgs.nc",
        help="Location to save stack of averaged igrams (default=%(default)s)",
    )
    p.add_argument(
        "--deramp-order",
        default=2,
        type=int,
        help="Specify order of surface to remove from phase when averaging. "
        " 1 = linear ramp, 2 = quadratic surface, 0 = no ramp adjustment (default=%(default)s)",
    )
    p.add_argument(
        "--nsigma",
        "-n",
        type=int,
        help=(
            "Number of sigma_mad deviations away from median to label as outlier"
            " (default=%(default)s)"
        ),
        default=5,
    )
    p.add_argument(
        "--rsc-file", help="If using ROI_PAC .rsc files, location of .rsc file"
    )
    p.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite existing averaged files (default=%(default)s)",
    )
    p.add_argument(
        "--max-temporal-baseline",
        type=int,
        help=(
            "Maximum temporal baseline to use when averaging interferograms. (default=%(default)s)"
        ),
        default=400,
    )
    p.add_argument(
        "--mask-files",
        nargs="+",
        help="List of binary mask files (e.g. water mask) to apply when creating average interferograms.",
    )
    p.add_argument(
        "--mask-is-zero",
        action="store_true",
        help=(
            "Indicate that the areas to mask are 0 in the files given by `--mask-files`. "
            "The default is the numpy convention where True (1) is masked."
        ),
    )
    p.add_argument(
        "--no-data",
        type=int,
        help="Indicate a nodata value to mask in all interferograms (default=%(default)s)",
    )
    p.add_argument(
        "--no-sign-flip",
        action="store_false",
        dest="do_flip",
        help="Skip the sign-flipping that makes interferogram averages have same direction."
        " Skipping for interferograms will make averages including long term deformation, "
        "but is useful for, e.g., averaging correlation images.",
    )
    return p.parse_args()


def average_and_label():
    """ """
    args = get_cli_args()
    core.create_averages(**vars(args))
    core.label_outliers(
        fname=args.avg_file,
        outfile=args.outfile,
        nsigma=args.nsigma,
        level=args.level,
    )



"""
This module exports a Log class that wraps the logging python package

Uses the standard python logging utilities, just provides
nice formatting out of the box.

Usage:

    from logger import get_log
    logger = get_log()

    logger.info("Something happened")
    logger.warning("Something concerning happened")
    logger.error("Something bad happened")
    logger.critical("Something just awful happened")
    logger.debug("Extra printing we often don't need to see.")
"""
import logging
import time
from functools import wraps
from logging import Formatter


def get_log(debug=False, name=__file__, verbose=False):
    """

    Parameters
    ----------
    debug : bool
         (Default value = False)
    name : str
         (Default value = __file__)
    verbose : bool
         (Default value = False)

    Returns
    -------

    """
    logger = logging.getLogger(name)
    return format_log(logger, debug=debug, verbose=verbose)


def format_log(logger, debug=False, verbose=False):
    log_level = logging.DEBUG if debug else logging.INFO

    format_ = "[%(asctime)s] [%(levelname)s %(filename)s] %(message)s"
    formatter = Formatter(format_, datefmt="%m/%d %H:%M:%S")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    if not logger.handlers:
        logger.addHandler(handler)
        logger.setLevel(log_level)

        if verbose:
            logger.info("Logger initialized: %s" % (logger.name,))

    if debug or verbose:
        logger.setLevel(logging.DEBUG)

    return logger


logger = get_log()


def log_runtime(f):
    """Logs how long a decorated function takes to run

    Parameters
    ----------
    f : function
        The function to wrap
        Retruns:
    function :
        The wrapped function
    """
    # pass function's docstring through
    @wraps(f)
    def wrapper(*args, **kwargs):
        """

        Parameters
        ----------
        *args :
            
        **kwargs :
            

        Returns
        -------

        """
        t1 = time.time()
        result = f(*args, **kwargs)
        t2 = time.time()
        elapsed_time = t2 - t1

        time_string = "Total elapsed time for {} : {} minutes ({} seconds)".format(
            f.__name__,
            "{0:.2f}".format(elapsed_time / 60.0),
            "{0:.2f}".format(elapsed_time),
        )
        logger.info(time_string)
        return result

    return wrapper



import numpy as np


def plot_avgs(fname="average_slcs.nc", stack=None, cmap=None, nimg=9):
    """Plot a fact grid of average interferograms.

    Parameters
    ----------
    fname : str
        Name of computed average interferogram file.
        Defaults to "average_slcs.nc".
    stack : xr.Dataset
        Alternative to `fname`, pass the average Dataset.
        Defaults to None.
    cmap : _type_
        Colormap for plotting. (Default value = None)
    nimg : int
        Number of total images to plot. Defaults to 9.

    Returns
    -------

    """
    import xarray as xr

    # import matplotlib.pyplot as plt
    if stack is None:
        with xr.open_dataarray(fname) as ds:
            stack = ds[:nimg]
    else:
        stack = stack[:nimg]

    # vmin, vmax = np.nanmin(avgs), np.nanmax(avgs)
    # vm = np.max(np.abs([vmin, vmax]))
    ntotal = stack.shape[0]
    ntiles = nimg if nimg < ntotal else ntotal

    nside = int(np.ceil(np.sqrt(ntiles)))
    stack.plot(
        x="lon",
        y="lat",
        col="date",
        col_wrap=nside,
        cmap=cmap,
        # vmax=np.percentile(ds.data, 95),
    )



#!/usr/bin/env python
"""
Create labels of outlier interferogram pixels.

Uses the averaged unwrapped igrams per date.
"""
import os

import numpy as np

from . import sario, utils
from .deramp import remove_ramp
from .logger import get_log, log_runtime

log = get_log()


def label_outliers(
    fname=None,
    stack=None,
    outfile="labels.nc",
    nsigma=5,
    level="pixel",
    min_spread=0.5,
):
    """

    Parameters
    ----------
    fname : str
        Filename of average interferogram (Default value = None)
    stack : xr.Dataset
        xr.Dataset containing the average, alternative to `fname` (Default value = None)
    outfile : str
        Name of output file to save (Default value = "labels.nc")
    nsigma : int
        Cutoff level to label outliers (Default value = 5)
    level : str
        Type of outlier labeling to use.
            "pixel" runs on each pixel individually
            "scene" takes each image variance as the input
            (Default value = "pixel")
    min_spread : float
        minimum value to use for calculating variances (Default value = 0.5)

    Returns
    -------
        labels, threshold: The labeled xr.Dataset and the threshold used to label outliers
    """
    if stack is None:
        import xarray as xr

        stack = xr.open_dataarray(fname)
    log.info("Computing {} sigma outlier labels at {} level.".format(nsigma, level))

    if level == "pixel":
        # Use all pixel absolute values here, shape: (ndates, rows, cols)
        stack_data = np.abs(stack)
        labels, threshold = label(stack_data, nsigma=nsigma, min_spread=min_spread)
    elif level == "scene":
        # Use just scene-level variance, shape: (ndates, 1, 1)
        stack_data = np.var(stack, axis=(1, 2), keepdims=True)
        labels, threshold = label(stack_data, nsigma=nsigma, min_spread=min_spread)
        # Add squeeze for the scene-level case, dont need lat/lon dims
        labels = labels.squeeze()
        stack_data = stack_data.squeeze()
        threshold = threshold.squeeze()
    else:
        raise ValueError("`level` must be 'pixel' or 'scene'")

    # Rename the xarray dataarrays
    labels = labels.rename("labels")
    stack_data = stack_data.rename("data")
    threshold = threshold.rename("threshold")
    if outfile:
        log.info("Saving outlier labels to {}:/labels".format(outfile))
        labels.to_netcdf(outfile)
        log.info("Saving data to {}:/data".format(outfile))
        stack_data.to_netcdf(outfile, mode="a")
        log.info("Saving threshold to {}:/threshold".format(outfile))
        threshold.to_netcdf(outfile, mode="a")
    return labels, threshold


def label(
    data,
    nsigma=5,
    min_spread=0.5,
):
    """Label outliers using the average interferograms

    Parameters
    ----------
    data : xr.DataArray (or np.ndarray)


    nsigma : int
         (Default value = 5)
    min_spread : float
         (Default value = 0.5)

    Returns
    -------
    labels : xr.DataArray
    threshold : xr.DataArray
    """
    med = data.median(axis=0)
    spread = np.maximum(min_spread, nsigma * mad(data, axis=0))
    threshold = med + spread
    return (data > threshold), threshold


def mad(stack, axis=0, scale=1.4826):
    """Median absolute deviation,

    Default `scale` is such that +/-MAD covers 50% (between 1/4 and 3/4)
    of the standard normal cumulative distribution

    Parameters
    ----------
    stack : xr.DataArray, or np.ndarray

    axis : int, optional
        axis along which to compute the MAD (Default value = 0)
    scale : float
        Multiplier to use for std dev (Default value = 1.4826)

    """
    stack_abs = np.abs(stack)
    med = np.nanmedian(stack_abs, axis=axis)
    return scale * np.nanmedian(np.abs(stack_abs - med), axis=axis)


@log_runtime
def create_averages(
    search_path=".",
    ext=".unw",
    rsc_file=None,
    deramp_order=2,
    avg_file="average_ifgs.nc",
    overwrite=False,
    band=2,
    ds_name="average_ifgs",
    max_temporal_baseline=800,
    do_flip=True,
    mask=None,
    mask_files=[],
    mask_is_zero=False,
    **kwargs,
):
    """Create a NetCDF stack of "average interferograms" for each date

    Parameters
    ----------
    search_path : str
        directory to find igrams (Default value = ".")
    ext : str
        extension name of unwrapped interferograms (default = .unw)
    rsc_file : str
        filename of .rsc resource file, if loading binary files like snaphu outputs (Default value = None)
    deramp_order : int
        remove a linear (or quadratic ramp) from unwrapped igrams
        if `deramp_order` = 1 (or 2) (Default value = 2)
    avg_file : str
        name of output file to save stack (Default value = "average_ifgs.nc")
    overwrite : bool
        clobber current output file, if exists (Default value = False)
    band : int
        if using rasterio to load igrams, which image band to load (Default value = 2)
    ds_name : str
        Name of the data variable used in the netcdf stack (Default value = "average_ifgs")
    do_flip : bool
        Flip the sign of interferograms to always go from (cur date, other date) (Default value = True)
    max_temporal_baseline : int
        Maximum temporal baseline to use for averaging, in days. (Default value = 800)
    mask : np.ndarray
        binary mask to apply to all loaded interferograms (Default value = None)
    mask_files : list
        List of files to use as masks for the stack. (Default value = [])
    mask_is_zero : bool
        If True, areas in the `mask_files` which are 0 are the places to ignore. (Default value = False.)
        Note that `mask_is_zero=False` means that True/1s are the mask areas (matching numpy masking defaults).

    Returns
    -------
    str: name of output file
    """
    import h5netcdf.legacyapi as nc

    if os.path.exists(avg_file) and not overwrite:
        log.info("{} exists, not overwriting.".format(avg_file))
        return avg_file

    log.info("Searching for igrams in {} with extention {}".format(search_path, ext))
    ifg_date_list = utils.find_igrams(directory=search_path, ext=ext)
    unw_file_list = utils.find_igrams(directory=search_path, ext=ext, parse=False)
    sar_date_list = utils.dates_from_igrams(ifg_date_list)

    nigrams, ndates = len(ifg_date_list), len(sar_date_list)
    log.info("Found {} igrams, {} unique SAR dates".format(nigrams, ndates))

    utils.create_empty_nc_stack(
        avg_file,
        date_list=sar_date_list,
        rsc_file=rsc_file,
        gdal_file=unw_file_list[0],
        stack_data_name=ds_name,
        overwrite=overwrite,
    )

    f = nc.Dataset(avg_file, mode="r+")
    ds = f[ds_name]
    _, rows, cols = ds.shape

    # Get masks for deramping
    # mask_igram_date_list = utils.load_intlist_from_h5(mask_fname)
    if mask is None:
        mask = np.zeros((rows, cols)).astype(bool)
    if mask_files:
        mask = np.logical_or(mask, sario.load_mask(mask_files, mask_is_zero=mask_is_zero))

    for (idx, cur_date) in enumerate(sar_date_list):
        cur_unws = [
            (fname, date_pair)
            for (date_pair, fname) in zip(ifg_date_list, unw_file_list)
            if (
                cur_date in date_pair
                and _temp_baseline(date_pair) <= max_temporal_baseline
            )
        ]
        log.info(
            "Averaging {} igrams for {} ({} out of {})".format(
                len(cur_unws), cur_date, idx + 1, len(sar_date_list)
            )
        )

        # reset the matrix to all zeros
        out = 0
        for unwf, date_pair in cur_unws:
            # Since each ifg of (date1, date2) was made by phase2 - phase1,
            # flip ifg phase so that it's always positive: (other date, cur_date)
            # otherwise the date's phase was negative in the interferogram
            flip = -1 if do_flip and (cur_date == date_pair[0]) else 1
            img = sario.load(unwf, rsc_file=rsc_file, band=band)
            out += flip * img

            # mask_idx = mask_igram_date_list.index(date_pair)
            # mask |= mask_stack[mask_idx]

        out /= len(cur_unws)

        if deramp_order > 0:
            out = remove_ramp(out, deramp_order=deramp_order, mask=mask)
        else:
            out -= out.mean()
            out[mask] = np.nan

        # Write the single layer out
        ds[idx, :, :] = out

    # Close to save it
    f.close()
    return avg_file


def _temp_baseline(date_pair):
    """Calculate the temporal baseline of a date pair"""
    return abs(date_pair[1] - date_pair[0]).days



# -*- coding: utf-8 -*-
"""
    Setup file for trodi.
    Use setup.cfg to configure your project.
"""
import sys
from pkg_resources import VersionConflict, require
from setuptools import setup

try:
    require("setuptools>=38.3")
except VersionConflict:
    print("Error: version of setuptools is too old (<38.3)!")
    sys.exit(1)

# try:
#     from Cython.Build import cythonize
# except ImportError:
#     import subprocess

#     subprocess.call([sys.executable, "-m", "pip", "install", "Cython>=0.27.1"])
#     from Cython.Build import cythonize


if __name__ == "__main__":
    setup()
